import Control.Monad(liftM, ap)

data Exp =
     Plus Exp Exp
   | Minus Exp Exp
   | Times Exp Exp
   | Div Exp Exp
   | Const Int


data Exn a = Ok a | Error String      
 deriving Show


andthen :: Exn a -> (a -> Exn b) -> Exn b
e `andthen` k =
  case e of
    Ok x    -> k x
    Error s -> Error s

----------------

{-

Monads, Formally
----------------


Formally, a monad is a type constructor 

  M a 

and two operations

  (>>=) :: M a -> (a -> M b) -> M b   
  return :: a -> M a

Note that we use the same names for the general case 
as for IO actions, and we continue to pronounce (>>=) as "bind".


The Monad type class
--------------------

The Prelude defines a class for monadic behavior:

  class Monad m where
    (>>=)  :: m a -> (a -> m b) -> m b
    return :: a -> m a

Unlike the other classes we have seen, this is
a TYPE CONSTRUCTOR class (m is a variable representing
a type constructor rather than a type).

Monad instances
---------------

The IO type constructor is declared as Monad instance,
using built-in primitive definitions roughly like this:

  instance Monad IO where
    (>>=)  = builtinBindIO
    return = builtinReturnIO

We can make Exn an instance:

-}


instance Monad Exn where
  (>>=)  = andthen
  return = Ok

{-

The "do" notation can be used for ANY
instance of the Monad class, including
user-defined instances.

So we can rewrite interpreter code using "do":

-}


eval :: Exp -> Exn Int
eval (Plus e1 e2)  =
     do v1 <- eval e1
        v2 <- eval e2
        return (v1+v2)
eval (Minus e1 e2) = 
     do v1 <- eval e1
        v2 <- eval e2
        return (v1-v2)
eval (Times e1 e2) = 
     do v1 <- eval e1
        v2 <- eval e2
        return (v1*v2)
eval (Div e1 e2)   =
     do v1 <- eval e1
        v2 <- eval e2
        if v2 == 0 then
          Error "divby0"
        else
          return (v1 `div` v2)
eval (Const i)     =  return i

example1 = eval (Div (Plus (Const 5) (Const 3))
                     (Const 2))

example2 = eval (Minus (Const 4) (Div (Const 2) (Const 0)))

{-

Monad Laws
----------

The monad operations should satisfy these three laws:

Associativity:

 (m >>= k) >>= h  ≡  m >>= (\x -> (k x >>= h))

Identities:

 return x >>= k ≡ k x

 m >>= return ≡ m


Notice that associativity is crucial for the
the "do" notation to make sense.

The Maybe Monad
---------------


In fact, the (very similar) Maybe type constructor is
already declared as a monad instance in the Prelude:

  data Maybe a = Just a | Nothing

  instance Monad Maybe where
    return = Just
    (Just x) >>= k = k x
    Nothing  >>= k = Nothing



Aside: The Functor type class
----------------------

Another important type constructor class in the Prelude is:

  class Functor f where
    fmap :: (a -> b) -> f a -> f b

Some standard instances:

  instance Functor [] where
    fmap = map
  instance Functor Maybe where
    fmap _ Nothing  = Nothing
    fmap g (Just x) = Just (g x)

We can also build instances for our own data types, e.g.

-}

data Tree a = Node (Tree a) (Tree a) | Leaf a
mapT :: (a -> b) -> Tree a -> Tree b
mapT f (Node l r) = Node (mapT f l) (mapT f r)
mapT f (Leaf x)   = Leaf (f x)     

instance Functor Tree where
  fmap = mapT
  
{-

Back to Monads...
--------------

Some dirty details:

In fact, Monad is a sub-class of Functor and of another constructor class
called Applicative, so we must also declared instances for them.

The following boilerplate should work for any monad:

import Control.Monad(liftM, ap)
instance Monad Foo where …
instance Functor Foo where
  fmap = liftM
instance Applicative Foo where
  pure = return
  (<*>) = ap

In particular, for Exn:

-}

instance Functor Exn where
  fmap = liftM
instance Applicative Exn where
  pure = return
  (<*>) = ap

  
{-

The textbook has much more to say about Applicative. 

Generic Monad Operations
------------------------

The operations we defined over IO actions are actually defined
for arbitrary monads, e.g.

  mapM :: Monad m => (a -> m b) -> [a] –> m [b]
  sequence :: Monad m => [m a] -> m [a]
  foldM: :: Monad m => (b -> a -> m b) -> b -> [a] -> m b

(foldM is a LEFT fold, i.e. a variant of foldl, not foldr)

-}

