import Control.Monad(liftM, ap)

data Exp =
     Plus Exp Exp
   | Minus Exp Exp
   | Times Exp Exp
   | Div Exp Exp
   | Const Int

-------------------


{- To clean up the program, let's introduce a new kind 
   monad for accumulating traces.

   The basic idea is that traces Tr will carry a trace string
   as well as a value.  The `return` operator injects a value
   into Tr with an empty trace; bind concatenates traces.
   We add an additional helper function `trace` to insert
   a string into the trace.

-}

data Tr a = Tr String a
     deriving Show

instance Monad Tr where
  return x = Tr "" x
  (Tr s x) >>= k =
     let Tr s' x' = k x
     in Tr (s++s') x'

trace :: String -> Tr ()
trace s = Tr s ()

runTr :: Tr a -> (String,a)      
runTr (Tr s x) = (s,x)
      
instance Functor Tr where
  fmap = liftM
instance Applicative Tr where
  pure = return
  (<*>) = ap

eval :: Exp -> Tr Int
eval (Plus e1 e2)  =
     do v1 <- eval e1
        v2 <- eval e2
        trace " +"
        return (v1 + v2)
eval (Minus e1 e2)  =
     do v1 <- eval e1
        v2 <- eval e2
        trace " -"
        return (v1 - v2)
eval (Times e1 e2)  =
     do v1 <- eval e1
        v2 <- eval e2
        trace " *"
        return (v1 * v2)
eval (Div e1 e2)  =
     do v1 <- eval e1
        v2 <- eval e2
        trace " /"
        return (v1 `div` v2)
eval (Const i)     =
     do trace (" " ++ show i)
        return i

example1 = runTr $ eval (Div (Plus (Const 5) (Const 3))
                             (Const 2))

{- A monad like Tr is sometimes called a "writer" monad, because
   the effect it describes is writing to a global. -}

   

