import Control.Monad(liftM, ap)

-------------------

{- Now let's extend our expression language to include variables.
   Variables are initially 0, and can be updated by assignment
   expressions.
-}

data Exp =
     Plus Exp Exp
   | Minus Exp Exp
   | Times Exp Exp
   | Div Exp Exp
   | Const Int
   | Var Id       -- yields current value of Id
   | (:=) Id Exp  -- assigns alue of Exp to Id and also yields that value

type Id = String

{-

To keep track of the values of variables, we'll use an environment
represented as a map similar to the one from the last homework:

-}

type Map a b = a -> Maybe b
empty :: Map a b 
empty = \x -> Nothing
update :: Eq a => Map a b -> a -> b -> Map a b
update m x v = \x' -> if x' == x then Just v else m x' 
fetch :: Map a b -> a -> Maybe b
fetch m x = m x 

type Env = Map Id Int

{-
	
We need to thread this Env through evaluation of the expression.
To do that, we could modify our original evaluator along
the following lines:


  eval :: Exp -> Env -> (Env,Exp)
  eval (Plus e1 e2) env =
     let (env1,v1) = eval e1 env
         (env2,v2) = eval e2 env1
     in (env2,v1+v2)
  ...
  eval (Var x) env = fetch env x
  eval (x := e) env = let (env',v) = eval env e in (update env' x v,v)

Instead of giving full details of this, let's jump write in to
writing a suitable "state" monad. 

-}

newtype SM a = SM (Env -> (Env,a))

instance Monad SM where
  return x = SM(\env -> (env,x))
  (SM m1) >>= k =
     SM(\env ->
           let (env',x) = m1 env in
           let SM m2 = k x in
               m2 env')

{- Monad-specific operations -}

get :: Id -> SM Int
get x = SM(\env -> (env,case fetch env x of {Just v -> v ; Nothing -> 0}))

put :: Id -> Int -> SM ()
put x v = SM(\env -> (update env x v,()))        

runSM :: SM Int -> Int 
runSM (SM m) = snd (m empty)
      
instance Functor SM where
  fmap = liftM
instance Applicative SM where
  pure = return
  (<*>) = ap

eval :: Exp -> SM Int
eval (Plus e1 e2)  =
     do v1 <- eval e1
        v2 <- eval e2
        return (v1 + v2)
eval (Minus e1 e2)  =
     do v1 <- eval e1
        v2 <- eval e2
        return (v1 - v2)
eval (Times e1 e2)  =
     do v1 <- eval e1
        v2 <- eval e2
        return (v1 * v2)
eval (Div e1 e2)  =
     do v1 <- eval e1
        v2 <- eval e2
        return (v1 `div` v2)
eval (Const i)     = return i
eval (Var x) =
     get x
eval (x := e) =
     do v <- eval e
        put x v
        return v

example1 = runSM $ eval (Div (Plus (Const 5) (Const 3))
                             (Const 2))

example2 = runSM $ eval (Plus (Plus (Const 1) ("x" := Const 2))
                        (Plus (Var "x") ("x" := Const 3)))

{- Notice that the code for `eval` essentially doesn't change
   as we add these different monadic effects, except for use
   of the "special-purpose" operations such as 'get' or 'trace'.

   An obvious question: can we combine all these effects into one
   version of eval that provides exceptions, tracing, and state?

   Answer: Sure!  There are (at least) two ways:

   (1) Define a new type constructor that incorporates	
       all three effects, e.g.

       data M a = M (Env -> Exn(Env,String,a))
   
       or, perhaps:

       data M a = M (Env -> (Env,String,Exn a))

       [Can you see the difference?]

       and define bind, return and all the special purpose operators
       appropriately.

   (2) Use "monad transformers" to build up the combination monad
       in a modular way.  This is cleaner but a bit more complicated.
   

-}
	